import logging
from functools import cached_property
from typing import Any

from banal import ensure_list
from followthemoney import EntityProxy, model

from openaleph_search.index.entities import ENTITY_SOURCE
from openaleph_search.index.indexes import entities_read_index, schema_bucket
from openaleph_search.index.mapping import Field
from openaleph_search.query.base import Query
from openaleph_search.query.matching import match_query
from openaleph_search.query.util import field_filter_query
from openaleph_search.settings import Settings

log = logging.getLogger(__name__)

EXCLUDE_SCHEMATA = [
    s.name for s in model.schemata.values() if s.hidden
]  # Page, Mention


class EntitiesQuery(Query):
    TEXT_FIELDS = [
        f"{Field.NAME}^4",
        f"{Field.NAMES}^3",
        f"{Field.NAME_PARTS}^2",
        Field.CONTENT,
        f"{Field.TEXT}^0.8",
    ]
    PREFIX_FIELD = Field.NAME_PARTS
    HIGHLIGHT_FIELD = Field.CONTENT
    SKIP_FILTERS = [Field.SCHEMA, Field.SCHEMATA]
    SOURCE = ENTITY_SOURCE
    SORT_DEFAULT = []

    @cached_property
    def schemata(self) -> list[str]:
        schemata = self.parser.getlist("filter:schema")
        if len(schemata):
            return schemata
        schemata = self.parser.getlist("filter:schemata")
        if not len(schemata):
            schemata = ["Thing"]
        return schemata

    def get_index(self):
        return entities_read_index(schema=self.schemata)

    def get_query(self) -> dict[str, Any]:
        return self.wrap_query_function_score(self.get_inner_query())

    def get_inner_query(self) -> dict[str, Any]:
        return super().get_query()

    def get_negative_filters(self) -> list[dict[str, Any]]:
        # exclude hidden schemata unless we explicitly want them
        filters = super().get_negative_filters()
        exclude_schemata = set(EXCLUDE_SCHEMATA) - set(self.schemata)
        filters.append(field_filter_query("schema", exclude_schemata))
        return filters

    def get_index_weight_functions(self) -> list[dict[str, Any]]:
        """Generate index weight functions based on schema bucket settings"""
        settings = Settings()
        functions = []

        # Map bucket names to their boost settings
        bucket_boosts = {
            "pages": settings.index_boost_pages,
            "documents": settings.index_boost_documents,
            "intervals": settings.index_boost_intervals,
            "things": settings.index_boost_things,
        }

        # Create boost functions for each schema in the query
        for schema_name in self.schemata:
            try:
                bucket = schema_bucket(schema_name)
                boost_value = bucket_boosts.get(bucket, 1)

                if boost_value != 1:  # Only add function if boost differs from default
                    functions.append(
                        {
                            "filter": {"term": {"schema": schema_name}},
                            "weight": boost_value,
                        }
                    )
            except (ValueError, KeyError):
                # Skip invalid schemas
                continue

        return functions

    def wrap_query_function_score(self, query: dict[str, Any]) -> dict[str, Any]:
        # Wrap query in function_score to up-score important entities.
        # (thank you, OpenSanctions/yente :))
        functions = [
            {
                "field_value_factor": {
                    "field": Field.NUM_VALUES,
                    # This is a bit of a jiggle factor. Currently, very
                    # large documents (like Vladimir Putin) have a
                    # num_values of ~200, so get a +10 boost.  The order
                    # is modifier(factor * value)
                    "factor": 0.5,
                    "modifier": "sqrt",
                }
            }
        ]

        # Add index weight functions
        functions.extend(self.get_index_weight_functions())

        return {
            "function_score": {
                "query": query,
                "functions": functions,
                "boost_mode": "sum",
            }
        }


class MatchQuery(EntitiesQuery):
    """Given an entity, find the most similar other entities."""

    def __init__(
        self,
        parser,
        entity: EntityProxy | None = None,
        exclude=None,
        datasets=None,
        collection_ids=None,
    ):
        self.entity = entity
        self.exclude = ensure_list(exclude)
        self.datasets = datasets
        self.collection_ids = collection_ids
        super(MatchQuery, self).__init__(parser)

    def get_index(self):
        # Attempt to find only matches within the "matchable" set of
        # entity schemata. For example, a Company and be matched to
        # another company or a LegalEntity, but not a Person.
        # Real estate is "unmatchable", i.e. even if two plots of land
        # have almost the same name and criteria, it does not make
        # sense to suggest they are the same.
        schemata = list(self.entity.schema.matchable_schemata)
        return entities_read_index(schema=schemata)

    def get_inner_query(self) -> dict[str, Any]:
        query = match_query(
            self.entity,
            datasets=self.datasets,
            collection_ids=self.collection_ids,
            query=super().get_inner_query(),
        )
        if len(self.exclude):
            exclude = {"ids": {"values": self.exclude}}
            query["bool"]["must_not"].append(exclude)
        return query


class GeoDistanceQuery(EntitiesQuery):
    """Given an Address entity, find the nearby Address entities via the
    geo_point field"""

    def __init__(self, parser, entity=None, exclude=None, datasets=None):
        self.entity = entity
        self.exclude = ensure_list(exclude)
        self.datasets = datasets
        super(EntitiesQuery, self).__init__(parser)

    def is_valid(self) -> bool:
        return (
            self.entity is not None
            and self.entity.first("latitude") is not None
            and self.entity.first("longitude") is not None
        )

    def get_query(self):
        if not self.is_valid():
            return {"match_none": {}}
        query = super(GeoDistanceQuery, self).get_query()
        exclude = {"ids": {"values": self.exclude + [self.entity.id]}}
        query["bool"]["must_not"].append(exclude)
        query["bool"]["must"].append({"exists": {"field": "geo_point"}})
        return query

    def get_sort(self):
        """Always sort by calculated distance"""
        if not self.is_valid():
            return []
        return [
            {
                "_geo_distance": {
                    "geo_point": {
                        "lat": self.entity.first("latitude"),
                        "lon": self.entity.first("longitude"),
                    },
                    "order": "asc",
                    "unit": "km",
                    "mode": "min",
                    "distance_type": "plane",  # faster
                }
            }
        ]



from anystore.logging import get_logger

from openaleph_search.core import get_es
from openaleph_search.index.indexes import (
    configure_entities,
    entities_read_index,
)

log = get_logger(__name__)


def upgrade_search():
    """Add any missing properties to the index mappings."""
    configure_entities()


def delete_index():
    es = get_es()
    log.warning("🔥 Deleting all indices 🔥")
    for index in entities_read_index().split(","):
        if es.indices.exists(index=index):
            es.indices.delete(index=index)


def clear_index():
    es = get_es()
    log.warning("🔥 Deleting all data 🔥")
    es.delete_by_query(
        index=entities_read_index(),
        body={"query": {"match_all": {}}},
        refresh=True,
        wait_for_completion=True,
        conflicts="proceed",
    )



import logging
from typing import Any, TypeAlias

from followthemoney import EntityProxy, Schema
from followthemoney.types import registry
from ftmq.util import get_name_symbols
from rigour.text import levenshtein

from openaleph_search.index.mapping import Field, property_field_name
from openaleph_search.query.util import BoolQuery, bool_query, none_query
from openaleph_search.transform.util import (
    index_name_keys,
    index_name_parts,
    phonetic_names,
    preprocess_name,
)

log = logging.getLogger(__name__)
Clauses: TypeAlias = list[dict[str, Any]]
MATCH_GROUPS = [
    registry.ip.group,
    registry.url.group,
    registry.email.group,
    registry.phone.group,
]
MAX_CLAUSES = 500


def pick_names(names: list[str], limit: int = 3) -> list[str]:
    """Try to pick a few non-overlapping names to search for when matching
    an entity. The problem here is that if we receive an API query for an
    entity with hundreds of aliases, it becomes prohibitively expensive to
    search. This function decides which ones should be queried as pars pro
    toto in the index before the Python comparison algo later checks all of
    them.

    This is a bit over the top and will come back to haunt us."""
    if len(names) <= limit:
        return names
    picked: list[str] = []
    processed_ = [preprocess_name(n) for n in names]
    names = [n for n in processed_ if n is not None]

    # Centroid:
    picked_name = registry.name.pick(names)
    if picked_name is not None:
        picked.append(picked_name)

    # Pick the least similar:
    for _ in range(1, limit):
        candidates: dict[str, int] = {}
        for cand in names:
            if cand in picked:
                continue
            candidates[cand] = 0
            for pick in picked:
                candidates[cand] += levenshtein(pick, cand)

        if not len(candidates):
            break
        pick, _ = sorted(candidates.items(), key=lambda c: c[1], reverse=True)[0]
        picked.append(pick)

    return picked


def names_query(schema: Schema, names: list[str]) -> Clauses:
    shoulds: Clauses = []
    for name in pick_names(names, limit=5):
        match = {
            Field.NAMES: {
                "query": name,
                "operator": "AND",
                "boost": 3.0,
                "fuzziness": "AUTO",
            }
        }
        shoulds.append({"match": match})
    for key in index_name_keys(schema, names):
        term = {Field.NAME_KEYS: {"value": key, "boost": 4.0}}
        shoulds.append({"term": term})
    for token in index_name_parts(schema, names):
        term = {Field.NAME_PARTS: {"value": token, "boost": 1.0}}
        shoulds.append({"term": term})
    for phoneme in phonetic_names(schema, names):
        term = {Field.NAME_PHONETIC: {"value": phoneme, "boost": 0.8}}
        shoulds.append({"term": term})
    for symbol in get_name_symbols(schema, *names):
        shoulds.append({"term": {Field.NAME_SYMBOLS: str(symbol)}})

    return shoulds


def identifiers_query(entity: EntityProxy) -> Clauses:
    shoulds: Clauses = []
    for prop, value in entity.itervalues():
        if prop.type.group == registry.identifier.group:
            term = {property_field_name(prop.name): {"value": value, "boost": 3.0}}
            shoulds.append({"term": term})
    return shoulds


def match_query(
    entity: EntityProxy,
    datasets: list[str] | None = None,
    collection_ids: list[str] | None = None,
    query: BoolQuery | None = None,
):
    """Given a matchable entity in indexed form, build a query that will find
    similar entities based on a variety of criteria. For other entities with
    more full text (e.g. documents), there is a "more_like_this" query in the
    `similar.py` query module"""

    if not entity.schema.matchable:
        return none_query()

    if query is None:
        query = bool_query()

    # Don't match the query entity
    must_not = []
    if entity.id is not None:
        must_not.append({"ids": {"values": [entity.id]}})
    if len(must_not):
        query["bool"]["must_not"].extend(must_not)

    if collection_ids:
        query["bool"]["filter"].append({"terms": {"collection_id": datasets}})
    elif datasets:
        query["bool"]["filter"].append({"terms": {"dataset": datasets}})

    # match on magic names
    names = entity.get_type_values(registry.name, matchable=True)
    names_lookup = names_query(entity.schema, names)
    if names_lookup:
        query["bool"]["must"].append(
            {"bool": {"should": names_lookup, "minimum_should_match": 1}}
        )

    # match on identifiers
    identifiers_lookup = identifiers_query(entity)
    if identifiers_lookup:
        query["bool"]["must"].append(
            {"bool": {"should": identifiers_lookup, "minimum_should_match": 0}}
        )

    # num clauses so far, if we have nothing, not useful to match at all
    num_clauses = len(names_lookup) + len(identifiers_lookup)
    if not num_clauses:
        return none_query()

    # match on other useful properties, sorted by specificity
    filters = set()
    for prop, value in entity.itervalues():
        specificity = prop.specificity(value)
        if specificity > 0:
            filters.add((prop.type, value, specificity))
    filters = sorted(filters, key=lambda p: p[2], reverse=True)
    groups = []
    for type_, value, _ in filters:
        if type_.group in MATCH_GROUPS and num_clauses <= MAX_CLAUSES:
            groups.append({"term": {type_.group: {"value": value, "boost": 2.0}}})
            num_clauses += 1

    scoring = []
    for type_, value, _ in filters:
        if type_.group not in MATCH_GROUPS and num_clauses <= MAX_CLAUSES:
            scoring.append({"term": {type_.group: {"value": value}}})
            num_clauses += 1

    query["bool"]["should"].extend(groups)
    query["bool"]["should"].extend(scoring)

    return query



import itertools
import unicodedata
from functools import lru_cache
from typing import List, Optional, Set

from anystore.logging import get_logger
from followthemoney import EntityProxy
from followthemoney.schema import Schema
from normality import ascii_text, collapse_spaces
from rigour.names import (
    remove_person_prefixes,
    replace_org_types_compare,
    tokenize_name,
)
from rigour.text import metaphone
from rigour.text.scripts import is_modern_alphabet

log = get_logger(__name__)


def get_geopoints(entity: EntityProxy) -> list[dict[str, str]]:
    """Get lon/lat pairs for indexing to `geo_point` field"""
    points = []
    lons = entity.get("longitude", quiet=True)
    lats = entity.get("latitude", quiet=True)
    for lon, lat in itertools.product(lons, lats):
        points.append({"lon": lon, "lat": lat})
    return points


def preprocess_name(name: Optional[str]) -> Optional[str]:
    """Preprocess a name for comparison."""
    if name is None:
        return None
    name = unicodedata.normalize("NFC", name)
    name = name.lower()
    return collapse_spaces(name)


@lru_cache(maxsize=2000)
def clean_tokenize_name(schema: Schema, name: str) -> List[str]:
    """Tokenize a name and clean it up."""
    name = preprocess_name(name) or name
    if schema.name in ("LegalEntity", "Organization", "Company", "PublicBody"):
        name = replace_org_types_compare(name, normalizer=preprocess_name)
    elif schema.name in ("LegalEntity", "Person"):
        name = remove_person_prefixes(name)
    return tokenize_name(name)


def phonetic_names(schema: Schema, names: List[str]) -> Set[str]:
    """Generate phonetic forms of the given names."""
    phonemes: Set[str] = set()
    for name in names:
        for token in clean_tokenize_name(schema, name):
            if len(token) < 3 or not is_modern_alphabet(token):
                continue
            if token.isnumeric():
                continue
            phoneme = metaphone(ascii_text(token))
            if len(phoneme) > 2:
                phonemes.add(phoneme)
    return phonemes


def index_name_parts(schema: Schema, names: List[str]) -> Set[str]:
    """Generate a list of indexable name parts from the given names."""
    parts: Set[str] = set()
    for name in names:
        for token in clean_tokenize_name(schema, name):
            if len(token) < 2:
                continue
            parts.add(token)
            # TODO: put name and company symbol lookups here
            if is_modern_alphabet(token):
                ascii_token = ascii_text(token)
                if ascii_token is not None and len(ascii_token) > 1:
                    parts.add(ascii_token)
    return parts


def index_name_keys(schema: Schema, names: List[str]) -> Set[str]:
    """Generate a indexable name keys from the given names."""
    keys: Set[str] = set()
    for name in names:
        tokens = clean_tokenize_name(schema, name)
        ascii_tokens: List[str] = []
        for token in tokens:
            if token.isnumeric() or not is_modern_alphabet(token):
                ascii_tokens.append(token)
                continue
            ascii_token = ascii_text(token) or token
            ascii_tokens.append(ascii_token)
        ascii_name = "".join(sorted(ascii_tokens))
        if len(ascii_name) > 5:
            keys.add(ascii_name)
    return keys



from typing import Any

from openaleph_search.index.mapping import Field


def get_highlighter(
    field: str, query: dict[str, Any] | None = None, count: int | None = None
) -> dict[str, Any]:
    # Content field - best UX with term vectors
    if field == Field.CONTENT:
        highlighter = {
            "type": "fvh",
            "fragment_size": 400,
            "fragment_offset": 50,
            "number_of_fragments": count or 3,
            "phrase_limit": 256,
            "order": "score",  # Best fragments first
            "boundary_scanner": "sentence",  # Break at sentences
            "boundary_max_scan": 50,
            # "pre_tags": ["<em class='highlight-content'>"],
            # "post_tags": ["</em>"],
            "max_analyzed_offset": 999999,  # Handle large documents
        }
        if query:
            highlighter["highlight_query"] = query
        return highlighter
    # Human-readable names - exact highlighting
    if field == Field.NAME:
        highlighter = {
            "type": "unified",  # Good for mixed content
            "fragment_size": 200,  # Longer to capture full names/titles
            "number_of_fragments": 3,
            "fragmenter": "simple",  # Don't break names awkwardly
            "max_analyzed_offset": 999999,  # Handle large documents
            "pre_tags": [""],  # No markup
            "post_tags": [""],  # No markup
            # "pre_tags": ["<em class='highlight-name'>"],
            # "post_tags": ["</em>"],
        }
        return highlighter
    # Keyword names - simple exact matching
    if field == Field.NAMES:
        return {
            "type": "plain",  # Fast for keyword fields
            "number_of_fragments": 3,
            "max_analyzed_offset": 1000,  # Limit for keyword fields
            "pre_tags": [""],  # No markup
            "post_tags": [""],  # No markup
        }
    # other fields - leftovers, minimal highlighting if possible (not important)
    plain = {
        "type": "plain",  # Fastest option
        "fragment_size": 150,  # Shorter since less important
        "number_of_fragments": 1,  # Just one fragment
        "max_analyzed_offset": 999999,  # Handle large documents
        # "pre_tags": ["<em class='highlight-text'>"],
        # "post_tags": ["</em>"],
    }
    if query:
        plain["highlight_query"] = query
    return plain



from typing import TypeAlias

from anystore.functools import weakref_cache
from followthemoney import Schema
from followthemoney.dataset.util import dataset_name_check

SchemaType: TypeAlias = Schema | str


@weakref_cache
def valid_dataset(dataset: str) -> str:
    return dataset_name_check(dataset)



from anystore.settings import BaseSettings
from pydantic import AliasChoices, Field, HttpUrl
from pydantic_settings import SettingsConfigDict

__version__ = "0.0.0"

MAX_PAGE = 9999
BULK_PAGE = 1000


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=".env", env_prefix="openaleph_", extra="ignore"
    )

    testing: bool = Field(
        default=False, validation_alias=AliasChoices("testing", "debug")
    )

    search_auth: bool = False
    """Set to true when using with OpenAleph"""

    search_auth_field: str = "dataset"
    """Default field to filter/apply auth on"""

    elasticsearch_uri: HttpUrl | list[HttpUrl] = HttpUrl("http://localhost:9200")
    elasticsearch_timeout: int = 60
    elasticsearch_max_retries: int = 3
    elasticsearch_retry_on_timeout: bool = True

    indexer_concurrency: int = 8
    indexer_chunk_size: int = 1000
    indexer_max_chunk_bytes: int = 50 * 1024 * 1024

    index_shards: int = 25  # 4 indices with dataset routing
    index_replicas: int = 0
    index_prefix: str = "openaleph"
    index_write: str = "v1"
    index_read: list[str] = ["v1"]
    index_expand_clause_limit: int = 10
    index_delete_by_query_batchsize: int = 100
    index_namespace_ids: bool = True

    # configure different weights for indices
    index_boost_intervals: int = 1
    index_boost_things: int = 1
    index_boost_documents: int = 1
    index_boost_pages: int = 1



"""High level search interface"""

from typing import Any
from urllib.parse import parse_qsl

from elastic_transport import ObjectApiResponse

from openaleph_search.core import get_es
from openaleph_search.index.indexes import entities_read_index
from openaleph_search.parse.parser import SearchQueryParser
from openaleph_search.query.queries import EntitiesQuery


def search_query_string(q: str, args: str | None = None) -> ObjectApiResponse:
    """Search using `query_string` with optional parser args"""
    _args = parse_qsl(args, keep_blank_values=True)
    if "q" in dict(_args):
        raise RuntimeError("Invalid query, must not contain `q` in args")
    if "highlight" not in dict(_args):
        _args.append(("highlight", "true"))
    _args.insert(0, ("q", q))
    parser = SearchQueryParser(_args)
    query = EntitiesQuery(parser)
    return query.search()


def search_body(
    body: dict[str, Any], index: str | None = None, routing: str | None = None
) -> ObjectApiResponse:
    es = get_es()
    index = index or entities_read_index()
    return es.search(index=index, body=body, routing=routing)



from anystore.logging import get_logger
from anystore.types import SDict
from banal import ensure_list

from openaleph_search.index.mapping import ANALYZE_SETTINGS
from openaleph_search.settings import Settings

log = get_logger(__name__)
settings = Settings()


def index_name(name: str, version: str) -> str:
    return "-".join((settings.index_prefix, name, version))


def check_response(index, res):
    """Check if a request succeeded."""
    if res.get("status", 0) > 399 and not res.get("acknowledged"):
        error = res.get("error", {}).get("reason")
        log.error("Index [%s] error: %s" % (index, error))
        return False
    return True


def refresh_sync(sync: bool | None = False) -> bool:
    return settings.testing or bool(sync)


def unpack_result(res: SDict) -> SDict | None:
    """Turn a document hit from ES into a more traditional JSON object."""
    error = res.get("error")
    if error is not None:
        raise RuntimeError("Query error: %r" % error)
    if res.get("found") is False:
        return
    data = res.get("_source", {})
    data["id"] = res.get("_id")
    data["_index"] = res.get("_index")
    data["_routing"] = res.get("_routing")

    _score = res.get("_score")
    if _score is not None and _score != 0.0 and "score" not in data:
        data["score"] = _score

    if "highlight" in res:
        data["highlight"] = res["highlight"]
        # data["highlight"] = []
        # for value in res.get("highlight", {}).values():
        #     data["highlight"].extend(value)

    data["_sort"] = ensure_list(res.get("sort"))
    return data


def check_settings_changed(updated, existing):
    """Since updating the settings requires closing the index, we don't
    want to do it unless it's really needed. This will check if all the
    updated settings are already in effect."""
    if not isinstance(updated, dict) or not isinstance(existing, dict):
        return updated != existing
    for key, value in list(updated.items()):
        if check_settings_changed(value, existing.get(key)):
            return True
    return False


def index_settings(
    shards: int | None = settings.index_shards,
    replicas: int | None = settings.index_replicas,
):
    """Configure an index in ES with support for text transliteration."""
    if settings.testing:
        shards = 1
        replicas = 0
    return {
        **ANALYZE_SETTINGS,
        "index": {
            "number_of_shards": str(shards),
            "number_of_replicas": str(replicas),
            "refresh_interval": "100ms" if settings.testing else "5s",
            "similarity": {
                # We use this for names, to avoid over-penalizing entities with many names.
                "weak_length_norm": {
                    # BM25 is the default similarity algorithm.
                    "type": "BM25",
                    # 0.75 is the default
                    "b": 0.25,
                }
            },
        },
    }



from anystore.decorators import error_handler
from anystore.functools import weakref_cache as cache
from anystore.logging import get_logger
from anystore.util import mask_uri
from banal import ensure_list
from elasticsearch import AsyncElasticsearch, Elasticsearch

from openaleph_search.settings import Settings

log = get_logger(__name__)


@cache
def _nodes() -> list[str]:
    nodes: set[str] = set()
    settings = Settings()
    for node in ensure_list(settings.elasticsearch_uri):
        nodes.add(str(node))
    return list(nodes)


@error_handler(logger=log)
def _get_client() -> Elasticsearch:
    settings = Settings()
    urls = _nodes()
    es = Elasticsearch(
        hosts=urls,
        request_timeout=settings.elasticsearch_timeout,
        max_retries=settings.elasticsearch_max_retries,
        retry_on_timeout=settings.elasticsearch_retry_on_timeout,
        retry_on_status=[502, 503, 504],
    )
    es.info()
    urls = [mask_uri(u) for u in urls]
    log.info("Connected to Elasticsearch", nodes=urls)
    return es


@error_handler(logger=log)
async def _get_async_client() -> AsyncElasticsearch:
    settings = Settings()
    urls = _nodes()
    es = AsyncElasticsearch(
        hosts=urls,
        request_timeout=settings.elasticsearch_timeout,
        max_retries=settings.elasticsearch_max_retries,
        retry_on_timeout=settings.elasticsearch_retry_on_timeout,
        retry_on_status=[502, 503, 504],
    )
    await es.info()
    log.info("Connected to AsyncElasticsearch", nodes=urls)
    return es


@cache
def get_es() -> Elasticsearch:
    return _get_client()


async def get_async_es() -> AsyncElasticsearch:
    return await _get_async_client()



from typing import Any

from pydantic import BaseModel

from openaleph_search.query.util import auth_datasets_query
from openaleph_search.settings import Settings

settings = Settings()


class SearchAuth(BaseModel):
    """Control auth for dataset filter"""

    datasets: set[str] = set()
    logged_in: bool = False
    is_admin: bool = False
    role: str | None = None

    # deprecated
    collection_ids: set[int] = set()

    def datasets_query(
        self, field: str | None = settings.search_auth_field
    ) -> dict[str, Any]:
        field = field or settings.search_auth_field
        if "collection" in field:
            return auth_datasets_query(
                list(map(str, self.collection_ids)), field, self.is_admin
            )
        return auth_datasets_query(list(self.datasets), field, self.is_admin)



"""Wrap cached rigour taggers loading into thread-safe calls"""

import importlib
import sys
import threading

import cachetools
from anystore.logging import get_logger
from rigour.names import normalize_name, tagging
from rigour.names.tagging import _get_org_tagger, _get_person_tagger

log = get_logger(__name__)
current_thread = threading.current_thread().ident

# Create shared caches and locks for each tagger
_person_tagger_cache = cachetools.LRUCache(maxsize=1)
_person_tagger_lock = threading.RLock()

_org_tagger_cache = cachetools.LRUCache(maxsize=1)
_org_tagger_lock = threading.RLock()

# Store original function implementations without their @cache decorators
original_person_tagger = (
    _get_person_tagger.__wrapped__
    if hasattr(_get_person_tagger, "__wrapped__")
    else _get_person_tagger
)
original_org_tagger = (
    _get_org_tagger.__wrapped__
    if hasattr(_get_org_tagger, "__wrapped__")
    else _get_org_tagger
)

PERSON_TAGGER = original_person_tagger(normalize_name)
ORG_TAGGER = original_org_tagger(normalize_name)


def patch_rigour_taggers():
    """Patch _get_person_tagger and _get_org_tagger for thread safety"""

    # Patch _get_person_tagger
    @cachetools.cached(cache=_person_tagger_cache, lock=_person_tagger_lock)
    def thread_safe_get_person_tagger(*args, **kwargs):
        log.info("Loading thread safe person tagger...", thread=current_thread)
        return PERSON_TAGGER

    tagging._get_person_tagger = thread_safe_get_person_tagger

    # Patch _get_org_tagger
    @cachetools.cached(cache=_org_tagger_cache, lock=_org_tagger_lock)
    def thread_safe_get_org_tagger(*args, **kwargs):
        log.info("Loading thread safe org tagger...", thread=current_thread)
        return ORG_TAGGER

    tagging._get_org_tagger = thread_safe_get_org_tagger


# Apply patches before importing
patch_rigour_taggers()

# Force reload the ftmq.util module that depend on the patched functions
if "ftmq.util" in sys.modules:
    importlib.reload(sys.modules["ftmq.util"])

from ftmq.util import get_name_symbols, get_symbols, select_symbols  # noqa: E402

__all__ = ["get_symbols", "get_name_symbols", "select_symbols"]



from functools import cached_property
from typing import Any, Iterator, cast

from anystore.logging import get_logger
from banal import as_bool
from followthemoney.util import sanitize_text
from werkzeug.datastructures import MultiDict, OrderedMultiDict

from openaleph_search.index.mapping import Field
from openaleph_search.model import SearchAuth
from openaleph_search.settings import MAX_PAGE, Settings
from openaleph_search.util import valid_dataset

settings = Settings()
log = get_logger(__name__)


class QueryParser:
    """Hold state for common query parameters."""

    SORT_ASC = "asc"
    SORT_DESC = "desc"
    SORT_DEFAULT = SORT_ASC
    SORTS = [SORT_ASC, SORT_DESC]

    def __init__(
        self,
        args: MultiDict | dict[str, Any],
        auth: SearchAuth | None = None,
        limit: int | None = None,
        max_limit: int | None = MAX_PAGE,
    ):
        if self.settings.search_auth and auth is None:
            raise RuntimeError("An auth object is required.")

        if not isinstance(args, MultiDict):
            args = OrderedMultiDict(args)
        self.args: MultiDict = cast(MultiDict, args)
        self.auth = auth
        self.offset = max(0, self.getint("offset", 0) or 0)
        if limit is None:
            limit = min(max_limit or MAX_PAGE, max(0, self.getint("limit", 20) or 20))
        self.limit = limit
        self.next_limit = self.getint("next_limit", limit)
        self.text = sanitize_text(self.get("q"))
        self.prefix = sanitize_text(self.get("prefix"))

        self.filters = self.prefixed_items("filter:")
        self.excludes = self.prefixed_items("exclude:")
        self.empties = self.prefixed_items("empty:")

    @property
    def page(self) -> int:
        if self.limit == 0:
            return 1
        return (self.offset // self.limit) + 1

    def prefixed_items(self, prefix: str) -> dict[str, set[str]]:
        items = {}
        for key in self.args.keys():
            if not key.startswith(prefix):
                continue
            name = key[len(prefix) :]
            items[name] = set(self.getlist(key))
        return items

    @property
    def sorts(self) -> list[tuple[str, str]]:
        sort = []
        for value in self.getlist("sort"):
            direction = self.SORT_DEFAULT
            if ":" in value:
                value, direction = value.rsplit(":", 1)
            if direction in self.SORTS:
                sort.append((value, direction))
        return sort

    @property
    def items(self) -> Iterator[tuple[str, str]]:
        for key, value in self.args.items(multi=True):
            if key in ("offset", "limit", "next_limit"):
                continue
            value = sanitize_text(value, encoding="utf-8")
            if value is not None:
                yield key, value

    def getlist(self, name: str, default: list[str] | None = None) -> list[str]:
        values = []
        for value in self.args.getlist(name):
            value = sanitize_text(value, encoding="utf-8")
            if value is not None:
                values.append(value)
        return values or (default or [])

    def get(self, name: str, default: str | None = None) -> str | None:
        for value in self.getlist(name):
            return value
        return default

    def getintlist(self, name: str, default: list[str] | None = None) -> list[int]:
        values = []
        for value in self.getlist(name, default=default):
            try:
                values.append(int(value))
            except (ValueError, TypeError):
                pass
        return values

    def getint(self, name: str, default: int | None = None) -> int | None:
        for value in self.getintlist(name):
            return value
        return default

    def getbool(self, name: str, default: bool = False) -> bool:
        return as_bool(self.get(name), default=default)

    def to_dict(self) -> dict[str, Any]:
        parser = {
            "text": self.text,
            "prefix": self.prefix,
            "offset": self.offset,
            "limit": self.limit,
            "filters": {key: list(val) for key, val in self.filters.items()},
            "sorts": self.sorts,
            "empties": {key: list(val) for key, val in self.empties.items()},
            "excludes": {key: list(val) for key, val in self.excludes.items()},
        }
        return parser

    @property
    def settings(self) -> Settings:
        # useful for test runtime to patch env vars
        if settings.testing:
            return Settings()
        return settings


class SearchQueryParser(QueryParser):
    """ElasticSearch-specific query parameters."""

    # Facets with known, limited cardinality:
    SMALL_FACETS = ("schema", "schemata", "dataset", "countries", "languages")

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(SearchQueryParser, self).__init__(*args, **kwargs)
        self.offset = min(MAX_PAGE, self.offset)
        if (self.limit + self.offset) > MAX_PAGE:
            self.limit = max(0, MAX_PAGE - self.offset)

        # Set of field names to facet by (i.e. include the count of distinct
        # values in the result set). These must match 'keyword' fields in the
        # index.
        self.facet_names = set(self.getlist("facet"))

        # Set of field names to use significant terms aggregation for
        self.facet_significant_names = set(self.getlist("facet_significant"))

        # Query to use for highlighting, defaults to the search query
        if self.get("highlight_text"):
            raise RuntimeError("Custom highlight text not supported")
        self.highlight_text = self.text
        # self.highlight_text = self.get("highlight_text", self.text)
        # Include highlighted fragments of matching text in the result.
        self.highlight = self.getbool("highlight", False)
        # self.highlight = self.highlight and self.highlight_text
        # Length of each snippet in characters
        if self.get("highlight_length"):
            raise RuntimeError("Custom highlight length not supported")
        self.highlight_length = 200
        # Number of snippets per document, 0 = return full document text.
        self.highlight_count = self.getint("highlight_count", 3)
        # By default, the maximum number of characters analyzed for a highlight
        # request is bounded by the value defined in the
        # index.highlight.max_analyzed_offset setting (1000000 by default),
        # and when the number of characters exceeds this limit an error is
        # returned. By setting `max_analyzed_offset` to a non-negative value
        # lower than `index.highlight.max_analyzed_offset`, the highlighting
        # stops at this defined maximum limit, and the rest of the text is not
        # processed, thus not highlighted and no error is returned.
        self.max_highlight_analyzed_offset = self.getint(
            "max_highlight_analyzed_offset", 999999
        )

    @cached_property
    def collection_ids(self) -> set[str]:
        collections = self.filters.get("collection_id", set())
        collections.update(self.filters.get("collections", set()))
        if self.auth and not self.auth.is_admin:
            collections = collections & set(self.auth.collection_ids)
        return collections

    @cached_property
    def datasets(self) -> set[str]:
        datasets = self.filters.get("dataset", set())
        datasets.update(self.filters.get("datasets", set()))
        if self.auth and not self.auth.is_admin:
            datasets = datasets & set(self.auth.datasets)
        return {valid_dataset(d) for d in datasets}

    @cached_property
    def routing_key(self) -> str | None:
        # get dataset routing if dataset filter is set
        if self.collection_ids:
            return ",".join(sorted(self.collection_ids))
        if self.datasets:
            return ",".join(sorted(self.datasets))

    def get_facet_size(self, name: str) -> int:
        """Number of distinct values to be included (i.e. top N)."""
        facet_size = self.getint("facet_size:%s" % name, 20) or 20
        # Added to mitigate a DDoS by scripted facet bots (2020-11-24):
        if self.auth:
            if not self.auth.logged_in and name not in self.SMALL_FACETS:
                facet_size = min(50, facet_size)
        return facet_size

    def get_facet_total(self, name: str) -> bool:
        """Flag to perform a count of the total number of distinct values."""
        if self.auth:
            if not self.auth.logged_in and name not in self.SMALL_FACETS:
                return False
        return self.getbool("facet_total:%s" % name, False)

    def get_facet_values(self, name: str) -> bool:
        """Flag to disable returning actual values (i.e. count only)."""
        # Added to mitigate a DDoS by scripted facet bots (2020-11-24):
        if self.get_facet_size(name) == 0:
            return False
        return self.getbool("facet_values:%s" % name, True)

    def get_facet_type(self, name: str) -> str | None:
        return self.get("facet_type:%s" % name)

    def get_facet_interval(self, name: str) -> str | None:
        """Interval to facet on when faceting on date properties

        See https://www.elastic.co/guide/en/elasticsearch/reference/current/search-aggregations-bucket-datehistogram-aggregation.html#calendar_intervals   # noqa: B950
        for available options for possible values
        """
        return self.get("facet_interval:%s" % name)

    def get_facet_significant(self, name: str) -> bool:
        """Flag to use significant terms aggregation for a facet."""
        return name in self.facet_significant_names

    def get_facet_significant_size(self, name: str) -> int:
        """Number of distinct values to be included for significant terms (i.e. top N)."""
        facet_size = self.getint("facet_significant_size:%s" % name, 20) or 20
        # Added to mitigate a DDoS by scripted facet bots (2020-11-24):
        if self.auth:
            if not self.auth.logged_in and name not in self.SMALL_FACETS:
                facet_size = min(50, facet_size)
        return facet_size

    def get_facet_significant_total(self, name: str) -> bool:
        """Flag to perform a count of the total number of distinct values for
        significant terms."""
        if self.auth:
            if not self.auth.logged_in and name not in self.SMALL_FACETS:
                return False
        return self.getbool("facet_significant_total:%s" % name, False)

    def get_facet_significant_values(self, name: str) -> bool:
        """Flag to disable returning actual values for significant terms (i.e. count only)."""
        # Added to mitigate a DDoS by scripted facet bots (2020-11-24):
        if self.get_facet_significant_size(name) == 0:
            return False
        return self.getbool("facet_significant_values:%s" % name, True)

    def get_facet_significant_type(self, name: str) -> str | None:
        return self.get("facet_significant_type:%s" % name)

    def get_facet_significant_text(self) -> str | None:
        """Field to use for significant text aggregation, or None if not specified."""
        field = self.get("facet_significant_text")
        return field or Field.CONTENT if field is not None else None

    def get_facet_significant_text_size(self) -> int:
        """Number of significant text terms to return."""
        return self.getint("facet_significant_text_size", 5) or 5

    def get_facet_significant_text_min_doc_count(self) -> int:
        """Minimum document count for significant text terms."""
        return self.getint("facet_significant_text_min_doc_count", 5) or 5

    def get_facet_significant_text_shard_size(self) -> int:
        """Shard size for significant text aggregation."""
        return self.getint("facet_significant_text_shard_size", 200) or 200

    def to_dict(self) -> dict[str, Any]:
        parser = super().to_dict()
        parser["facet_names"] = list(self.facet_names)
        parser["facet_significant_names"] = list(self.facet_significant_names)
        return parser



import logging
from typing import Any, Literal, TypeAlias

from anystore.functools import weakref_cache as cache
from banal import ensure_list
from followthemoney import model
from followthemoney.exc import InvalidData
from followthemoney.schema import Schema

from openaleph_search.index.indexer import configure_index
from openaleph_search.index.mapping import Field, make_mapping, make_schema_mapping
from openaleph_search.index.util import index_name, index_settings
from openaleph_search.settings import Settings
from openaleph_search.util import SchemaType

log = logging.getLogger(__name__)
settings = Settings()

Bucket: TypeAlias = Literal["pages", "documents", "intervals", "things"]
BUCKETS = ("pages", "documents", "intervals", "things")


@cache
def ensure_schema(schema: SchemaType) -> Schema:
    schema_ = model.get(schema)
    if schema_ is not None:
        return schema_
    raise ValueError(f"Invalid schema: `{schema}`")


@cache
def schema_bucket(schema: SchemaType) -> Bucket:
    """Convert a schema to its index bucket"""
    schema = ensure_schema(schema)
    if schema.name in ("Page", "Pages"):
        return "pages"
    if schema.is_a("Document"):
        return "documents"
    if schema.is_a("Thing"):  # catch "Event"
        return "things"
    if schema.is_a("Interval"):
        return "intervals"
    return "things"  # FIXME e.g. Mentions


@cache
def bucket_index(bucket: Bucket, version: str):
    """Convert a bucket str to an index name."""
    name = "entity-%s" % bucket
    return index_name(name, version=version)


@cache
def schema_index(schema: SchemaType, version: str):
    """Convert a schema object to an index name."""
    schema = ensure_schema(schema)
    if schema.abstract:
        raise InvalidData("Cannot index abstract schema: %s" % schema)
    return bucket_index(schema_bucket(schema), version)


def schema_scope(
    schema: SchemaType | list[SchemaType] | None = None, expand: bool | None = True
):
    schemata: set[Schema] = set()
    for schema_ in ensure_list(schema) or model.schemata.values():
        if schema_:
            schema_ = ensure_schema(schema_)
            schemata.add(schema_)
            if expand:
                schemata.update(schema_.descendants)
    for schema in schemata:
        if not schema.abstract:
            yield schema


def entities_index_list(
    schema: SchemaType | list[SchemaType] | None = None, expand: bool | None = True
) -> set[str]:
    """Combined index to run all queries against."""
    indexes: set[str] = set()
    for schema_ in schema_scope(schema, expand=expand):
        for version in settings.index_read:
            indexes.add(schema_index(schema_, version))
    return indexes


def entities_read_index(
    schema: SchemaType | list[SchemaType] | None = None, expand: bool | None = True
) -> str:
    """Current configured read indexes"""
    indexes = entities_index_list(schema=schema, expand=expand)
    return ",".join(indexes)


def entities_write_index(schema):
    """Index that is currently written by new queries."""
    return schema_index(schema, settings.index_write)


def make_schema_bucket_properties(bucket: Bucket) -> dict[str, Any]:
    """Configure the property mapping for the given schema bucket"""
    schemata: set[Schema] = set()
    for schema in model.schemata.values():
        if schema_bucket(schema) == bucket:
            schemata.add(schema)
    return make_schema_mapping(schemata)


def configure_entities():
    """Configure all the entity indexes"""
    for bucket in BUCKETS:
        for version in settings.index_read:
            configure_schema_bucket(bucket, version)


def make_schema_bucket_mapping(bucket: Bucket) -> dict[str, Any]:
    properties = make_schema_bucket_properties(bucket)
    mapping = make_mapping(properties)
    if bucket == "pages":
        # store full text for highlighting
        mapping["properties"][Field.CONTENT]["store"] = True
    return mapping


def configure_schema_bucket(bucket: Bucket, version: str):
    """
    Generate relevant type mappings for entity properties so that
    we can do correct searches on each.
    """
    mapping = make_schema_bucket_mapping(bucket)
    index = bucket_index(bucket, version)
    settings = index_settings()
    return configure_index(index, mapping, settings)



from openaleph_search.index.util import unpack_result
from openaleph_search.parse.parser import QueryParser, SearchQueryParser
from openaleph_search.query.base import Query
from openaleph_search.query.queries import EntitiesQuery, GeoDistanceQuery, MatchQuery

__all__ = [
    "EntitiesQuery",
    "GeoDistanceQuery",
    "MatchQuery",
    "Query",
    "QueryParser",
    "SearchQueryParser",
    "unpack_result",
]


__version__ = "0.0.3"



from typing import Any, ClassVar

from anystore.logging import get_logger
from banal import ensure_list
from elastic_transport import ObjectApiResponse
from followthemoney.types import registry

from openaleph_search.core import get_es
from openaleph_search.index.mapping import (
    DATE_FORMAT,
    GROUPS,
    NUMERIC_TYPES,
    Field,
    get_field_type,
    get_index_field_type,
)
from openaleph_search.parse.parser import SearchQueryParser
from openaleph_search.query.highlight import get_highlighter
from openaleph_search.query.util import (
    BoolQuery,
    bool_query,
    field_filter_query,
    filter_text,
    range_filter_query,
)
from openaleph_search.settings import Settings

log = get_logger(__name__)
settings = Settings()


class Query:
    TEXT_FIELDS: ClassVar[list[str]] = [Field.TEXT]
    PREFIX_FIELD: ClassVar[str] = Field.NAME
    SKIP_FILTERS: ClassVar[list[str]] = []
    AUTHZ_FIELD: ClassVar[str | None] = settings.search_auth_field
    HIGHLIGHT_FIELD: ClassVar[str] = Field.TEXT
    SORT_FIELDS: ClassVar[dict[str, str]] = {
        "label": "label.kw",
        "score": "_score",
    }
    SORT_DEFAULT: ClassVar[list[str | dict[str, Any]]] = ["_score"]
    SOURCE: ClassVar[dict[str, Any]] = {}

    def __init__(self, parser: SearchQueryParser) -> None:
        self.parser = parser

    def get_query_string(self) -> dict[str, Any] | None:
        if self.parser.text:
            return {
                "query_string": {
                    "query": self.parser.text,
                    "lenient": True,
                    "fields": self.TEXT_FIELDS,
                    "default_operator": "AND",
                    "minimum_should_match": "66%",
                }
            }

    def get_text_query(self) -> list[dict[str, Any]]:
        query = []
        if self.parser.text:
            query.append(self.get_query_string())
        if self.parser.prefix:
            query.append({"prefix": {self.PREFIX_FIELD: self.parser.prefix}})
        if not len(query):
            query.append({"match_all": {}})
        return query

    def get_filters_list(self, skip: set[str]) -> list[dict[str, Any]]:
        filters = []
        range_filters = dict()
        for field, values in self.parser.filters.items():
            if field in skip:
                continue
            # Collect all range query filters for a field in a single query
            if field.startswith(("gt:", "gte:", "lt:", "lte:")):
                op, field = field.split(":", 1)
                if range_filters.get(field) is None:
                    range_filters[field] = {op: list(values)[0]}
                else:
                    range_filters[field][op] = list(values)[0]
                continue
            filters.append(field_filter_query(field, values))

        for field, ops in range_filters.items():
            filters.append(range_filter_query(field, ops))

        return filters

    def get_filters(self) -> list[dict[str, Any]]:
        """Apply query filters from the user interface."""
        skip = {*self.SKIP_FILTERS, *self.parser.facet_names}
        # important as we don't have schema indexes anymore:
        skip.discard("schema")
        skip.discard("schemata")

        filters = self.get_filters_list(skip)

        if self.AUTHZ_FIELD is not None:
            # This enforces the authorization (access control) rules on
            # a particular query by comparing the datasets a user is
            # authorized for with the one on the document.
            if self.parser.auth and not self.parser.auth.is_admin:
                datasets = self.parser.auth.datasets_query(self.AUTHZ_FIELD)
                filters.append(datasets)
        return filters

    def get_post_filters(self, exclude: str | None = None) -> dict[str, dict[str, Any]]:
        """Apply post-aggregation query filters."""
        pre = set(self.parser.filters.keys())
        pre = pre.difference(self.parser.facet_names)
        skip = {*pre, *self.SKIP_FILTERS, exclude}
        filters = self.get_filters_list(skip)
        return {"bool": {"filter": filters}}

    def get_negative_filters(self) -> list[dict[str, Any]]:
        """Apply negative filters."""
        filters = []
        for field, _ in self.parser.empties.items():
            filters.append({"exists": {"field": field}})

        for field, values in self.parser.excludes.items():
            filters.append(field_filter_query(field, values))
        return filters

    def get_query(self) -> dict[str, Any]:
        return {
            "bool": {
                "should": self.get_text_query(),
                "must": [],
                "must_not": self.get_negative_filters(),
                "filter": self.get_filters(),
                "minimum_should_match": 1,
            }
        }

    def get_aggregations(self) -> dict[str, Any]:
        """Aggregate the query in order to generate faceted results."""
        aggregations = {}

        # Regular facet aggregations
        for facet_name in self.parser.facet_names:
            facet_aggregations = {}
            if self.parser.get_facet_values(facet_name):
                agg_name = "%s.values" % facet_name
                terms = {
                    "field": facet_name,
                    "size": self.parser.get_facet_size(facet_name),
                    "execution_hint": "map",
                }
                facet_aggregations[agg_name] = {"terms": terms}

            if self.parser.get_facet_total(facet_name):
                # Option to return total distinct value counts for
                # a given facet, instead of the top buckets.
                agg_name = "%s.cardinality" % facet_name
                facet_aggregations[agg_name] = {"cardinality": {"field": facet_name}}

            interval = self.parser.get_facet_interval(facet_name)
            if interval is not None:
                agg_name = "%s.intervals" % facet_name
                facet_aggregations[agg_name] = {
                    "date_histogram": {
                        "field": facet_name,
                        "calendar_interval": interval,
                        "format": DATE_FORMAT,
                        "min_doc_count": 0,
                    }
                }
                # Make sure we return empty buckets in the whole filter range
                filters = self.parser.filters
                min_val = filters.get("gte:%s" % facet_name) or filters.get(
                    "gt:%s" % facet_name
                )
                max_val = filters.get("lte:%s" % facet_name) or filters.get(
                    "lt:%s" % facet_name
                )
                if min_val or max_val:
                    extended_bounds = {}
                    if min_val:
                        extended_bounds["min"] = ensure_list(min_val)[0]
                    if max_val:
                        extended_bounds["max"] = ensure_list(max_val)[0]
                    facet_aggregations[agg_name]["date_histogram"][
                        "extended_bounds"
                    ] = extended_bounds

            if len(facet_aggregations):
                # See here for an explanation of the whole post_filters and
                # aggregation filters thing:
                # https://www.elastic.co/guide/en/elasticsearch/reference/6.2/search-request-post-filter.html  # noqa: B950
                other_filters = self.get_post_filters(exclude=facet_name)
                if len(other_filters["bool"]["filter"]):
                    agg_name = "%s.filtered" % facet_name
                    aggregations[agg_name] = {
                        "filter": other_filters,
                        "aggregations": facet_aggregations,
                    }
                else:
                    aggregations.update(facet_aggregations)

        # Significant terms aggregations
        for facet_name in self.parser.facet_significant_names:
            facet_aggregations = {}
            if self.parser.get_facet_significant_values(facet_name):
                agg_name = "%s.significant_terms" % facet_name
                significant_terms_agg = {
                    "field": facet_name,
                    "background_filter": self.get_significant_background(),
                    "size": self.parser.get_facet_significant_size(facet_name),
                    "min_doc_count": 3,
                    "shard_size": max(
                        100, self.parser.get_facet_significant_size(facet_name) * 5
                    ),
                    "execution_hint": "map",
                }
                facet_aggregations[agg_name] = {
                    "significant_terms": significant_terms_agg
                }

                if self.parser.get_facet_significant_type(facet_name) == "nested":
                    facet_aggregations[agg_name]["aggregations"] = {
                        agg_name: {"significant_terms": significant_terms_agg}
                    }

            if self.parser.get_facet_significant_total(facet_name):
                # Option to return total distinct value counts for significant terms
                agg_name = "%s.significant_cardinality" % facet_name
                facet_aggregations[agg_name] = {"cardinality": {"field": facet_name}}

            if len(facet_aggregations):
                # Apply post-filters for significant terms aggregations
                other_filters = self.get_post_filters(exclude=facet_name)
                if len(other_filters["bool"]["filter"]):
                    agg_name = "%s.significant_filtered" % facet_name
                    aggregations[agg_name] = {
                        "filter": other_filters,
                        "aggregations": facet_aggregations,
                    }
                else:
                    aggregations.update(facet_aggregations)

        significant_text_field = self.parser.get_facet_significant_text()
        if significant_text_field:
            aggregations["significant_text"] = {
                **self.get_significant_text_sampler(),
                "aggs": {
                    "significant_text": {
                        "significant_text": {
                            "field": significant_text_field,
                            "background_filter": self.get_significant_background(),
                            "filter_duplicate_text": True,
                            "size": self.parser.get_facet_significant_text_size(),
                            "min_doc_count": self.parser.get_facet_significant_text_min_doc_count(),  # noqa: B950
                            "shard_size": self.parser.get_facet_significant_text_shard_size(),
                        }
                    }
                },
            }

        return aggregations

    def get_significant_background(self) -> BoolQuery | None:
        query = bool_query()
        if self.parser.collection_ids:
            query["bool"]["must"].append(
                field_filter_query(Field.COLLECTION_ID, self.parser.collection_ids)
            )
        elif self.parser.datasets:
            query["bool"]["must"].append(
                field_filter_query(Field.DATASET, self.parser.datasets)
            )
        return query

    def get_sample_for_aggregation(
        self, size: int, agg_name: str, aggregation: dict[str, Any]
    ) -> dict[str, Any]:
        # https://www.elastic.co/docs/reference/aggregations/search-aggregations-bucket-sampler-aggregation
        return {"sampler": {"shard_size": size}, "aggs": {agg_name: aggregation}}

    def get_significant_text_sampler(self) -> dict[str, Any]:
        if self.parser.collection_ids or self.parser.datasets:
            # no sampling on all datasets
            return {"sampler": {"shard_size": 200}}
        return {"diversified_sampler": {"shard_size": 200, "field": self.AUTHZ_FIELD}}

    def get_sort(self) -> list[str | dict[str, dict[str, Any]]]:
        """Pick one of a set of named result orderings."""
        if not len(self.parser.sorts):
            return self.SORT_DEFAULT

        sort_fields = ["_score"]
        for field, direction in self.parser.sorts:
            field = self.SORT_FIELDS.get(field, field)
            type_ = get_field_type(field)
            config = {"order": direction, "missing": "_last"}
            es_type = get_index_field_type(type_, to_numeric=True)
            if es_type:
                config["unmapped_type"] = es_type
            if field == registry.date.group:
                field = "numeric.dates"
                config["mode"] = "min"
            if type_ in NUMERIC_TYPES:
                field = field.replace("properties.", "numeric.")
            sort_fields.append({field: config})
        return list(reversed(sort_fields))

    def get_highlight(self) -> dict[str, Any]:
        if not self.parser.highlight:
            return {}
        query = self.get_query_string()
        if self.parser.filters:
            query = bool_query()
            if self.get_query_string():
                query["bool"]["should"] = [self.get_query_string()]
            for key, values in self.parser.filters.items():
                if key in GROUPS or key == Field.NAME:
                    for value in values:
                        query["bool"]["should"].append(
                            {
                                "multi_match": {
                                    "fields": [Field.CONTENT, Field.TEXT, Field.NAME],
                                    "query": value,
                                    "operator": "AND",
                                }
                            }
                        )
        return {
            "encoder": "html",
            # "max_fragment_length": 1000,
            "require_field_match": False,
            "fields": {
                self.HIGHLIGHT_FIELD: get_highlighter(
                    self.HIGHLIGHT_FIELD, query, self.parser.highlight_count
                ),
                Field.NAMES: get_highlighter(Field.NAME),
                Field.NAMES: get_highlighter(Field.NAMES),
                Field.TEXT: get_highlighter(Field.TEXT, query),
            },
        }

    def get_source(self) -> dict[str, Any]:
        return self.SOURCE

    def has_significant_aggregations(self) -> bool:
        """Check if any significant aggregations are being performed."""
        # Check for significant_text aggregation
        if self.parser.get_facet_significant_text():
            return True

        # Check for significant terms aggregations on facets
        if self.parser.facet_significant_names:
            return True

        return False

    def get_body(self) -> dict[str, Any]:
        # Don't return hits when doing significant aggregations
        size = 0 if self.has_significant_aggregations() else self.parser.limit

        body = {
            "query": self.get_query(),
            "post_filter": self.get_post_filters(),
            "from": self.parser.offset,
            "size": size,
            "aggregations": self.get_aggregations(),
            "sort": self.get_sort(),
            "highlight": self.get_highlight(),
            "_source": self.get_source(),
        }
        # log.info("Query: %s", pformat(body))
        return body

    def get_index(self) -> str:
        raise NotImplementedError

    def to_text(self, empty: str = "*:*") -> str:
        """Generate a string representation of the query."""
        parts = []
        if self.parser.text:
            parts.append(self.parser.text)
        elif self.parser.prefix:
            query = "%s:%s*" % (self.PREFIX_FIELD, self.parser.prefix)
            parts.append(query)
        else:
            parts.append(empty)

        for filter_ in self.get_filters_list([]):
            if filter_.get("term", {}).get("schemata") == "Thing":
                continue
            parts.append(filter_text(filter_))

        for filter_ in self.get_negative_filters():
            parts.append(filter_text(filter_, invert=True))

        if len(parts) > 1 and empty in parts:
            parts.remove(empty)
        return " ".join([p for p in parts if p is not None])

    def search(self) -> ObjectApiResponse:
        """Execute the query as assmbled."""
        log.debug("Search index: %s" % self.get_index())
        es = get_es()
        result = es.search(
            index=self.get_index(),
            body=self.get_body(),
            routing=self.parser.routing_key,
        )
        log.debug(
            f"Elasticsearch query [{self.to_text()}] took {result.get('took')}ms",
            query=self.to_text(),
            routing=self.parser.routing_key,
            body=self.get_body(),
            filters=self.parser.filters,
            took=result.get("took"),
            hits=result.get("hits", {}).get("total", {}).get("value"),
        )
        return result



from typing import Any, Iterable, TypedDict

from anystore.types import SDict
from banal import ensure_list, is_mapping

from openaleph_search.index.mapping import Field
from openaleph_search.settings import Settings

settings = Settings()


class Bools(TypedDict):
    should: list[Any]
    filter: list[Any]
    must: list[Any]
    must_not: list[Any]


class BoolQuery(TypedDict):
    bool: Bools


def bool_query() -> BoolQuery:
    return {"bool": {"should": [], "filter": [], "must": [], "must_not": []}}


def none_query(query: BoolQuery | None = None) -> BoolQuery:
    if query is None:
        query = bool_query()
    query["bool"]["must"].append({"match_none": {}})
    return query


def field_filter_query(field: str, values: str | Iterable[str]) -> SDict:
    """Need to define work-around for full-text fields."""
    values = ensure_list(values)
    if not len(values):
        return {"match_all": {}}
    if field in ["_id", "id"]:
        return {"ids": {"values": values}}
    if field in ["names"]:
        field = Field.NAMES
    if len(values) == 1:
        # if field in ['addresses']:
        #     field = '%s.text' % field
        #     return {'match_phrase': {field: values[0]}}
        return {"term": {field: values[0]}}
    return {"terms": {field: values}}


def range_filter_query(field: str, ops: dict[str, Any]) -> SDict:
    return {"range": {field: ops}}


def filter_text(spec: Any, invert: bool = False) -> str | None:
    """Try to convert a given filter to a lucene query string."""
    # CAVEAT: This doesn't cover all filters used by aleph.
    if isinstance(spec, (list, tuple, set)):
        parts = [filter_text(s, invert=invert) for s in spec]
        return " ".join(p for p in parts if p)
    if not is_mapping(spec):
        return spec
    for op, props in spec.items():
        if op == "term":
            field, value = next(iter(props.items()))
            field = "-%s" % field if invert else field
            return '%s:"%s"' % (field, value)
        if op == "terms":
            field, values = next(iter(props.items()))
            parts = [{"term": {field: v}} for v in values]
            parts = [filter_text(p, invert=invert) for p in parts]
            predicate = " AND " if invert else " OR "
            text = predicate.join(p for p in parts if p)
            if len(parts) > 1:
                text = "(%s)" % text
            return text
        if op == "exists":
            field = props.get("field")
            field = "-%s" % field if invert else field
            return "%s:*" % field
    return None


def auth_datasets_query(
    values: list[str],
    field: str | None = settings.search_auth_field,
    is_admin: bool | None = False,
) -> dict[str, Any]:
    """Generate a search query filter for the given datasets."""
    # Hot-wire authorization entirely for admins.
    if is_admin:
        return {"match_all": {}}
    if not len(values):
        return {"match_none": {}}
    return {"terms": {field: values}}



from countrytagger import tag_place
from ftmq.util import get_name_symbols
from followthemoney import model
from functools import cached_property
from pydantic import BaseModel, computed_field


PERSON = model["Person"]
ORG = model["Organization"]


class QueryString(BaseModel):
    q: str

    @computed_field
    @cached_property
    def symbols(self) -> set[str]:
        symbols: set[str] = set()
        symbols.update(get_name_symbols(PERSON, [self.q]))
        symbols.update(get_name_symbols(ORG, [self.q]))
        return symbols

    @computed_field
    @cached_property
    def countries(self) -> set[str]:
        countries: set[str] = set()
        results = tag_place(self.q)
        if results is not None and len(results):
            countries = {r[2] for r in results}
        return countries



import logging
from typing import Any, Iterable

from banal import ensure_list
from elasticsearch.helpers import scan
from followthemoney import model
from followthemoney.proxy import EntityProxy
from followthemoney.types import registry

from openaleph_search.core import get_es
from openaleph_search.index.indexer import (
    MAX_REQUEST_TIMEOUT,
    MAX_TIMEOUT,
    bulk_actions,
    delete_safe,
)
from openaleph_search.index.indexes import (
    entities_read_index,
    entities_write_index,
)
from openaleph_search.index.util import unpack_result
from openaleph_search.model import SearchAuth
from openaleph_search.settings import MAX_PAGE
from openaleph_search.transform.entity import format_parallel

log = logging.getLogger(__name__)
PROXY_INCLUDES = [
    "caption",
    "schema",
    "properties",
    "dataset",
    "collection_id",
    "profile_id",
    "role_id",
    "mutable",
    "created_at",
    "updated_at",
]
ENTITY_SOURCE = {"includes": PROXY_INCLUDES}


def _source_spec(includes, excludes):
    includes = ensure_list(includes)
    excludes = ensure_list(excludes)
    return {"includes": includes, "excludes": excludes}


def _entities_query(
    filters: list[Any],
    auth: SearchAuth | None = None,
    dataset: str | None = None,
    collection_id: str | None = None,
    schemata: set[str] | None = None,
):
    filters = filters or []
    if auth is not None:
        filters.append(auth.datasets_query())
    if collection_id is not None:
        filters.append({"term": {"collection_id": collection_id}})
    elif dataset is not None:
        filters.append({"term": {"dataset": dataset}})
    if ensure_list(schemata):
        filters.append({"terms": {"schemata": ensure_list(schemata)}})
    return {"bool": {"filter": filters}}


def iter_entities(
    auth: SearchAuth | None = None,
    dataset: str | None = None,
    collection_id: str | None = None,
    schemata=None,
    includes=PROXY_INCLUDES,
    excludes=None,
    filters=None,
    sort=None,
    es_scroll="5m",
    es_scroll_size=1000,
):
    """Scan all entities matching the given criteria."""
    query = {
        "query": _entities_query(filters, auth, dataset, collection_id, schemata),
        "_source": _source_spec(includes, excludes),
    }
    preserve_order = False
    if sort is not None:
        query["sort"] = ensure_list(sort)
        preserve_order = True
    index = entities_read_index(schema=schemata)
    es = get_es()
    for res in scan(
        es,
        index=index,
        query=query,
        timeout=MAX_TIMEOUT,
        request_timeout=MAX_REQUEST_TIMEOUT,
        preserve_order=preserve_order,
        scroll=es_scroll,
        size=es_scroll_size,
        routing=collection_id,
    ):
        entity = unpack_result(res)
        if entity is not None:
            yield entity


def iter_proxies(**kw):
    for data in iter_entities(**kw):
        schema = model.get(data.get("schema"))
        if schema is None:
            continue
        yield model.get_proxy(data)


def iter_adjacent(dataset, entity_id):
    """Used for recursively deleting entities and their linked associations."""
    yield from iter_entities(
        includes=["dataset"],
        dataset=dataset,
        filters=[{"term": {"entities": entity_id}}],
    )


def entities_by_ids(
    ids, schemata=None, cached=False, includes=PROXY_INCLUDES, excludes=None
):
    """Iterate over unpacked entities based on a search for the given
    entity IDs."""
    ids = ensure_list(ids)
    if not len(ids):
        return
    if cached:
        # raise RuntimeError("Caching not implemented")
        log.warning("Caching not implemented")

    index = entities_read_index(schema=schemata)

    query = {
        "query": {"ids": {"values": ids}},
        "_source": _source_spec(includes, excludes),
        "size": MAX_PAGE,
    }
    es = get_es()
    result = es.search(index=index, body=query)
    for doc in result.get("hits", {}).get("hits", []):
        entity = unpack_result(doc)
        if entity is not None:
            yield entity


def get_entity(entity_id, **kwargs):
    """Fetch an entity from the index."""
    for entity in entities_by_ids(entity_id, cached=False, **kwargs):
        return entity


def index_proxy(dataset: str, proxy: EntityProxy, sync=False, **kwargs):
    delete_entity(proxy.id, exclude=proxy.schema, sync=False)
    return index_bulk(dataset, [proxy], sync=sync, **kwargs)


def index_bulk(dataset: str, entities: Iterable[EntityProxy], sync=False, **kwargs):
    """Index a set of entities."""
    actions = format_parallel(dataset, entities, **kwargs)
    bulk_actions(actions, sync=sync)


def delete_entity(entity_id, exclude=None, sync=False):
    """Delete an entity from the index."""
    if exclude is not None:
        exclude = entities_write_index(exclude)
    for entity in entities_by_ids(entity_id, excludes="*"):
        index = entity.get("_index")
        if index == exclude:
            continue
        delete_safe(index, entity_id, routing=entity.get("_routing"))


def checksums_count(checksums):
    """Query how many documents mention a checksum."""
    schemata = model.get_type_schemata(registry.checksum)
    index = entities_read_index(schemata)
    body = []
    for checksum in checksums:
        body.append({"index": index})
        query = {"term": {registry.checksum.group: checksum}}
        body.append({"size": 0, "query": query})
    es = get_es()
    results = es.msearch(body=body)
    for checksum, result in zip(checksums, results.get("responses", [])):
        total = result.get("hits", {}).get("total", {}).get("value", 0)
        yield checksum, total
